@@ -1,17 +1,18 @@
 #!/usr/bin/env python
+import time
+
 __author__ = 'rohe0002'
 
 import requests
 import random
 import string
-import copy
 import cookielib
-import logging
 from Cookie import SimpleCookie
 
 from oic.utils.keyio import KeyJar
 from oic.utils.time_util import utc_time_sans_frac
 from oic.exception import UnSupported
+import logging
 
 logger = logging.getLogger(__name__)
 
@@ -64,6 +65,10 @@ class TimeFormatError(Exception):
     pass
 
 
+class CapabilitiesMisMatch(Exception):
+    pass
+
+
 def rndstr(size=16):
     """
     Returns a string of random ascii characters or digits
@@ -260,22 +265,6 @@ PAIRS = {
     "path": "path_specified"
 }
 
-import time
-
-
-def _since_epoch(cdate):
-    # date format 'Wed, 06-Jun-2012 01:34:34 GMT'
-    try:
-        _cdate = cdate[5:-4]
-        try:
-            t = time.strptime(_cdate, "%d-%b-%Y %H:%M:%S")
-        except ValueError:
-            t = time.strptime(_cdate, "%d-%b-%y %H:%M:%S")
-    except Exception:
-        raise TimeFormatError(cdate)
-
-    return int(time.mktime(t))
-
 
 class PBase(object):
     def __init__(self, ca_certs=None, verify_ssl=True):
@@ -304,7 +293,9 @@ class PBase(object):
 
     def set_cookie(self, kaka):
         """PLaces a cookie (a cookielib.Cookie based on a set-cookie header
-        line) in the cookie jar. """
+        line) in the cookie jar.
+        Always chose the shortest expires time.
+        """
 
         # default rfc2109=False
         # max-age, httponly
@@ -325,12 +316,12 @@ class PBase(object):
                     if attr in ATTRS:
                         if morsel[attr]:
                             if attr == "expires":
-                                std_attr[attr] = _since_epoch(morsel[attr])
+                                std_attr[attr] = cookielib.http2time(morsel[attr])
                             else:
                                 std_attr[attr] = morsel[attr]
                     elif attr == "max-age":
                         if morsel[attr]:
-                            std_attr["expires"] = _since_epoch(morsel[attr])
+                            std_attr["expires"] = cookielib.http2time(morsel[attr])
             except TimeFormatError:
                 # Ignore cookie
                 logger.info(
@@ -374,7 +365,15 @@ class PBase(object):
         if self.cookiejar:
             _kwargs["cookies"] = self._cookies()
             logger.debug("SENT COOKIEs: %s" % (_kwargs["cookies"],))
-        r = requests.request(method, url, **_kwargs)
+
+        try:
+            r = requests.request(method, url, **_kwargs)
+        except Exception as err:
+            logger.error(
+                "http_request failed: %s, url: %s, htargs: %s, method: %s" % (
+                err, url, _kwargs, method))
+            raise
+
         try:
             set_cookie = r.headers["set-cookie"]
             # Telekom fix
@@ -424,7 +423,7 @@ class Client(PBase):
         self.verify_ssl = verify_ssl
         #self.secret_type = "basic "
 
-        self.state = None
+        #self.state = None
         self.nonce = None
 
         self.grant = {}
@@ -464,7 +463,7 @@ class Client(PBase):
     client_secret = property(get_client_secret, set_client_secret)
 
     def reset(self):
-        self.state = None
+        #self.state = None
         self.nonce = None
 
         self.grant = {}
@@ -517,18 +516,18 @@ class Client(PBase):
 
         return uri
 
-    def get_grant(self, **kwargs):
-        try:
-            _state = kwargs["state"]
-            if not _state:
-                _state = self.state
-        except KeyError:
-            _state = self.state
+    def get_grant(self, state, **kwargs):
+        # try:
+        #     _state = kwargs["state"]
+        #     if not _state:
+        #         _state = self.state
+        # except KeyError:
+        #     _state = self.state
 
         try:
-            return self.grant[_state]
+            return self.grant[state]
         except:
-            raise Exception("No grant found for state:'%s'" % _state)
+            raise Exception("No grant found for state:'%s'" % state)
 
     def get_token(self, also_expired=False, **kwargs):
         try:
@@ -667,8 +666,9 @@ class Client(PBase):
         request_args["access_token"] = token.access_token
         return self.construct_request(request, request_args, extra_args)
 
-    def get_or_post(self, uri, method, req,
-                    content_type=DEFAULT_POST_CONTENT_TYPE, accept=None, **kwargs):
+    @staticmethod
+    def get_or_post(uri, method, req, content_type=DEFAULT_POST_CONTENT_TYPE,
+                    accept=None, **kwargs):
         if method == "GET":
             _qp = req.to_urlencoded()
             if _qp:
@@ -687,7 +687,7 @@ class Client(PBase):
                     "Unsupported content type: '%s'" % content_type)
 
             header_ext = {"Content-type": content_type}
-            if (accept):
+            if accept:
                 header_ext = {"Accept": accept}
             
             if "headers" in kwargs.keys():
@@ -759,7 +759,7 @@ class Client(PBase):
         :param info: The response, can be either in a JSON or an urlencoded
             format
         :param sformat: Which serialization that was used
-        :param state:
+        :param state: The state
         :param kwargs: Extra key word arguments
         :return: The parsed and to some extend verified response
         """
@@ -850,33 +850,53 @@ class Client(PBase):
         else:
             return http_args
 
+    @staticmethod
+    def verify_header(reqresp, body_type):
+        logger.debug("resp.headers: %s" % (reqresp.headers,))
+        logger.debug("resp.txt: %s" % (reqresp.text,))
+
+        # This might be a tad to strict
+        if body_type == "":
+            pass
+        elif body_type == "json":
+            try:
+                assert "application/json" in reqresp.headers["content-type"]
+            except AssertionError:
+                try:
+                    assert "application/jwt" in reqresp.headers[
+                        "content-type"]
+                    body_type = "jwt"
+                except AssertionError:
+                    raise AssertionError("Wrong content-type in header")
+        elif body_type == "urlencoded":
+            try:
+                assert DEFAULT_POST_CONTENT_TYPE in reqresp.headers[
+                    "content-type"]
+            except AssertionError:
+                assert "text/plain" in reqresp.headers["content-type"]
+        else:
+            raise ValueError("Unknown return format: %s" % body_type)
+
+        return body_type
+
     def parse_request_response(self, reqresp, response, body_type, state="",
                                **kwargs):
 
         if reqresp.status_code in [200, 201]:
-            logger.debug("resp.headers: %s" % (reqresp.headers,))
-            logger.debug("resp.txt: %s" % (reqresp.text,))
-            if body_type == "":
-                pass
-            elif body_type == "json":
-                assert "application/json" in reqresp.headers["content-type"]
-            elif body_type == "urlencoded":
-                try:
-                    assert DEFAULT_POST_CONTENT_TYPE in reqresp.headers[
-                        "content-type"]
-                except AssertionError:
-                    assert "text/plain" in reqresp.headers["content-type"]
-            else:
-                raise ValueError("Unknown return format: %s" % body_type)
+            body_type = self.verify_header(reqresp, body_type)
         elif reqresp.status_code == 302:  # redirect
             pass
         elif reqresp.status_code == 500:
             logger.error("(%d) %s" % (reqresp.status_code, reqresp.text))
             raise Exception("ERROR: Something went wrong: %s" % reqresp.text)
+        elif reqresp.status_code == 400:
+            #expecting an error response
+            if issubclass(response, ErrorResponse):
+                pass
         else:
             logger.error("(%d) %s" % (reqresp.status_code, reqresp.text))
-            raise Exception("ERROR: Something went wrong: %s [%s]" % (
-                reqresp.text, reqresp.status_code))
+            raise Exception("HTTP ERROR: %s [%s] on %s" % (
+                reqresp.text, reqresp.status_code, reqresp.url))
 
         if body_type:
             if response:
@@ -922,6 +942,9 @@ class Client(PBase):
                                  response_cls=AuthorizationResponse,
                                  **kwargs):
 
+        if state:
+            request_args["state"] = state
+
         url, body, ht_args, csi = self.request_info(request, method,
                                                     request_args, extra_args,
                                                     **kwargs)
