@@ -1,3 +1,4 @@
+import copy
 import logging
 import urllib
 import urlparse
@@ -146,8 +147,11 @@ class Message(object):
                     for item in val:
                         params.append((key, str((unicode(item)).encode('utf-8'))))
             elif isinstance(val, Message):
-                params.append((key, str(_ser(val, sformat="urlencoded",
+                try:
+                    params.append((key, str(_ser(val, sformat="urlencoded",
                                              lev=lev))))
+                except TypeError:
+                    params.append((key, val))
             elif val is None:
                 params.append((key, val))
             else:
@@ -361,8 +365,9 @@ class Message(object):
                 else:
                     for v in val:
                         if not isinstance(v, vtype):
-                            raise DecodeError(ERRTXT % (key,
-                                                        "type != %s" % vtype))
+                            raise DecodeError(
+                                ERRTXT % (key, "type != %s (%s)" % (
+                                    vtype, type(v))))
 
                 self._dict[skey] = val
             else:
@@ -391,11 +396,11 @@ class Message(object):
                 else:
                     self._dict[skey] = val
 
-    def to_json(self, lev=0):
+    def to_json(self, lev=0, indent=None):
         if lev:
             return self.to_dict(lev + 1)
         else:
-            return json.dumps(self.to_dict(1))
+            return json.dumps(self.to_dict(1), indent=indent)
 
     def from_json(self, txt, **kwargs):
         return self.from_dict(json.loads(txt))
@@ -552,6 +557,11 @@ class Message(object):
                 if required:
                     raise MissingRequiredAttribute("%s" % attribute)
                 continue
+            else:
+                if not val:
+                    if required:
+                        raise MissingRequiredAttribute("%s" % attribute)
+                    continue
 
             if attribute not in _allowed:
                 continue
@@ -662,6 +672,9 @@ class Message(object):
         _res = jwe.decrypt(msg, krs)
         return self.from_json(_res[0])
 
+    def copy(self):
+            return copy.deepcopy(self)
+
 # =============================================================================
 
 
@@ -764,7 +777,7 @@ class AuthorizationErrorResponse(ErrorResponse):
     c_param.update({"state": SINGLE_OPTIONAL_STRING})
     c_allowed_values = ErrorResponse.c_allowed_values.copy()
     c_allowed_values.update({"error": ["invalid_request",
-                                       "unathorized_client",
+                                       "unauthorized_client",
                                        "access_denied",
                                        "unsupported_response_type",
                                        "invalid_scope", "server_error",
