@@ -1,3 +1,4 @@
+from jwkest.jwe import JWE
 from oic.utils.keyio import KeyJar
 
 __author__ = 'rohe0002'
@@ -50,7 +51,8 @@ from oic.utils.webfinger import OIC_ISSUER
 from oic.utils.webfinger import WebFinger
 
 from jwkest import jws
-from jwkest.jws import alg2keytype, JWS
+from jwkest import jwe
+from jwkest.jws import JWS
 
 logger = logging.getLogger(__name__)
 
@@ -202,7 +204,7 @@ PREFERENCE2PROVIDER = {
         "request_object_encryption_enc_values_supported",
     "userinfo_signed_response_alg": "userinfo_signing_alg_values_supported",
     "userinfo_encrypted_response_alg":
-        "userinfo_encryption_enc_values_supported",
+        "userinfo_encryption_alg_values_supported",
     "userinfo_encrypted_response_enc":
         "userinfo_encryption_enc_values_supported",
     "id_token_signed_response_alg": "id_token_signing_alg_values_supported",
@@ -220,6 +222,8 @@ PREFERENCE2PROVIDER = {
     #"request_object_signing_alg": "request_object_signing_alg_values_supported
 }
 
+PROVIDER2PREFERENCE = dict([(v,k) for k,v in PREFERENCE2PROVIDER.items()])
+
 PROVIDER_DEFAULT = {
     "token_endpoint_auth_method": "client_secret_basic",
     "id_token_signed_response_alg": "RS256",
@@ -300,6 +304,32 @@ class Client(oauth2.Client):
 
         return None
 
+    def request_object_encryption(self, msg, **kwargs):
+        try:
+            encalg = self.behaviour["request_object_encryption_alg"]
+        except KeyError:
+            return msg
+        else:
+            encenc = self.behaviour["request_object_encryption_enc"]
+            _jwe = JWE(msg, alg=encalg, enc=encenc)
+            _kty = jwe.alg2keytype(encalg)
+
+            try:
+                _kid = kwargs["enc_kid"]
+            except KeyError:
+                try:
+                    _kid = self.kid["enc"][_kty]
+                except KeyError:
+                    _kid = ""
+
+            if _kid:
+                _jwe["keys"] = self.keyjar.get_encrypt_key(_kty, kid=_kid)
+                _jwe["kid"] = _kid
+            else:
+                _jwe["keys"] = self.keyjar.get_signing_key(_kty)
+
+        return _jwe.encrypt(self.keyjar)
+
     def construct_AuthorizationRequest(self, request=AuthorizationRequest,
                                        request_args=None, extra_args=None,
                                        request_param=None, **kwargs):
@@ -321,7 +351,9 @@ class Client(oauth2.Client):
         if "request_method" in kwargs:
             if kwargs["request_method"] == "file":
                 request_param = "request_uri"
-                del kwargs["request_method"]
+            else:
+                request_param = "request"
+            del kwargs["request_method"]
 
         areq = oauth2.Client.construct_AuthorizationRequest(self, request,
                                                             request_args,
@@ -329,12 +361,16 @@ class Client(oauth2.Client):
                                                             **kwargs)
 
         if request_param:
-            alg = self.behaviour["request_object_signing_alg"]
-            if "algorithm" not in kwargs:
-                kwargs["algorithm"] = alg
+            try:
+                alg = self.behaviour["request_object_signing_alg"]
+            except KeyError:
+                alg = None
+            else:
+                if "algorithm" not in kwargs:
+                    kwargs["algorithm"] = alg
 
             if "keys" not in kwargs and alg:
-                _kty = alg2keytype(alg)
+                _kty = jws.alg2keytype(alg)
                 try:
                     kwargs["keys"] = self.keyjar.get_signing_key(
                         _kty, kid=self.kid["sig"][_kty])
@@ -343,12 +379,15 @@ class Client(oauth2.Client):
 
             _req = make_openid_request(areq, **kwargs)
 
+            # Should the request be encrypted
+            _req = self.request_object_encryption(_req, **kwargs)
+
             if request_param == "request":
                 areq["request"] = _req
             else:
                 _filedir = kwargs["local_dir"]
                 _webpath = kwargs["base_path"]
-                _name = rndstr(10)
+                _name = rndstr(10) + ".jwt"
                 filename = os.path.join(_filedir, _name)
                 while os.path.exists(filename):
                     _name = rndstr(10)
@@ -709,7 +748,7 @@ class Client(oauth2.Client):
             return _schema().from_json(txt=resp.text)
         else:
             algo = self.client_prefs["userinfo_signed_response_alg"]
-            _kty = alg2keytype(algo)
+            _kty = jws.alg2keytype(algo)
             # Keys of the OP ?
             try:
                 keys = self.keyjar.get_signing_key(_kty, self.kid["sig"][_kty])
@@ -942,21 +981,17 @@ class Client(oauth2.Client):
                     self.behaviour[_pref] = vals
             else:
                 vtyp = regreq.c_param[_pref]
-                if isinstance(vtyp[0], list):
-                    _list = True
-                else:
-                    _list = False
 
-                for val in vals:
-                    if val in _pvals:
-                        if not _list:
+                if isinstance(vtyp[0], list):
+                    self.behaviour[_pref] = []
+                    for val in vals:
+                        if val in _pvals:
+                            self.behaviour[_pref].append(val)
+                else:
+                    for val in vals:
+                        if val in _pvals:
                             self.behaviour[_pref] = val
                             break
-                        else:
-                            try:
-                                self.behaviour[_pref].append(val)
-                            except KeyError:
-                                self.behaviour[_pref] = [val]
 
             if _pref not in self.behaviour:
                 raise ConfigurationError(
@@ -1188,38 +1223,63 @@ class Server(oauth2.Server):
     #    def parse_issuer_request(self, info, sformat="urlencoded"):
     #        return self._parse_request(IssuerRequest, info, sformat)
 
+
     @staticmethod
-    def id_token_claims(session):
+    def update_claims(session, where, about, old_claims=None):
+        """
+
+        :param session:
+        :param where: Which request
+        :param about: userinfo or id_token
+        :param old_claims:
+        :return: claims or None
+        """
+
+        if old_claims is None:
+            old_claims = {}
+
+        req = None
+        if where == "oidreq":
+            try:
+                req = OpenIDRequest().deserialize(session[where], "json")
+            except KeyError:
+                pass
+        else:  # where == "authzreq"
+            try:
+                req = AuthorizationRequest().deserialize(session[where], "json")
+            except KeyError:
+                pass
+
+        if req:
+            logger.debug("%s: %s" % (where, req.to_dict()))
+            try:
+                _claims = req["claims"][about]
+                if _claims:
+                    # update with old claims, do not overwrite
+                    for key, val in old_claims.items():
+                        if key not in _claims:
+                            _claims[key] = val
+                    return _claims
+            except KeyError:
+                pass
+
+        return old_claims
+
+    def id_token_claims(self, session):
         """
         Pick the IdToken claims from the request
 
         :param session: Session information
         :return: The IdToken claims
         """
-        itc = None
-
-        try:
-            authzreq = AuthorizationRequest().deserialize(session["authzreq"],
-                                                          'json')
-            itc = authzreq["claims"]["id_token"]
-            logger.debug("ID Token claims: %s" % itc)
-        except KeyError:
-            pass
-
-        try:
-            oidreq = OpenIDRequest().deserialize(session["oidreq"], "json")
-            itc_or = oidreq["claims"]["id_token"]
-            if itc:
-                itc.update(itc_or)
-            logger.debug("ID Token claims: %s" % itc)
-        except KeyError:
-            pass
-
+        itc = {}
+        itc = self.update_claims(session, "authzreq", "id_token", itc)
+        itc = self.update_claims(session, "oidreq", "id_token", itc)
         return itc
 
     def make_id_token(self, session, loa="2", issuer="",
                       alg="RS256", code=None, access_token=None,
-                      user_info=None, auth_time=0):
+                      user_info=None, auth_time=0, exp=None):
         """
 
         :param session: Session information
@@ -1232,7 +1292,10 @@ class Server(oauth2.Server):
         :return: IDToken instance
         """
         #defaults
-        inawhile = {"days": 1}
+        if exp is None:
+            inawhile = {"days": 1}
+        else:
+            inawhile = exp
         # Handle the idtoken_claims
         extra = {}
         itc = self.id_token_claims(session)
@@ -1247,6 +1310,11 @@ class Server(oauth2.Server):
                 elif key == "acr":
                     #["2","http://id.incommon.org/assurance/bronze"]
                     extra["acr"] = verify_acr_level(val, loa)
+        else:
+            if auth_time:
+                extra["auth_time"] = auth_time
+            if loa:
+                extra["acr"] = loa
 
         if user_info is None:
             _args = {}
