@@ -70,6 +70,8 @@ def address_deser(val, sformat="urlencoded"):
         if not isinstance(val, basestring):
             val = json.dumps(val)
             sformat = "json"
+        elif sformat == "dict":
+            sformat = "json"
     return AddressClaim().deserialize(val, sformat)
 
 
@@ -457,14 +459,17 @@ class OpenIDSchema(Message):
 
     def verify(self, **kwargs):
         if "birthdate" in self:
-            # Either YYYY-MM-DD or just YYYY
+            # Either YYYY-MM-DD or just YYYY or 0000-MM-DD
             try:
                 _ = time.strptime(self["birthdate"], "%Y-%m-%d")
             except ValueError:
                 try:
                     _ = time.strptime(self["birthdate"], "%Y")
                 except ValueError:
-                    raise VerificationError("Birthdate format error", self)
+                    try:
+                        _ = time.strptime(self["birthdate"], "0000-%m-%d")
+                    except ValueError:
+                        raise VerificationError("Birthdate format error", self)
 
         return super(OpenIDSchema, self).verify(**kwargs)
 
@@ -514,6 +519,18 @@ class RegistrationRequest(Message):
         if "initiate_login_uri" in self:
             assert self["initiate_login_uri"].startswith("https:")
 
+        for param in ["request_object", "id_token", "userinfo"]:
+            if "%s_encryption_alg" % param in self:
+                if "%s_encryption_enc" % param not in self:
+                    self["%s_encryption_enc" % param] = "A128CBC-HS256"
+
+            # both or none
+            if "%s_encryption_enc" % param in self:
+                assert "%s_encryption_alg" % param in self
+
+        if "token_endpoint_auth_signing_alg" in self:
+            assert self["token_endpoint_auth_signing_alg"] != "none"
+
         return super(RegistrationRequest, self).verify(**kwargs)
 
 
@@ -554,7 +571,8 @@ class RegistrationResponse(Message):
 
 class ClientRegistrationErrorResponse(message.ErrorResponse):
     c_allowed_values = {"error": ["invalid_redirect_uri",
-                                  "invalid_client_metadata"]}
+                                  "invalid_client_metadata",
+                                  "invalid_configuration_parameter"]}
 
 
 class IdToken(OpenIDSchema):
