@@ -1,16 +1,15 @@
 #!/usr/bin/env python
-import copy
 import json
 import traceback
 import urllib
 import sys
 from jwkest.jwe import JWE
-from oic.utils import time_util
 from oic.utils.authn.user import NoSuchAuthentication
 from oic.utils.authn.user import ToOld
 from oic.utils.authn.user import TamperAllert
+from oic.utils.sdb import AuthnEvent
 from oic.utils.time_util import utc_time_sans_frac
-from oic.utils.keyio import KeyBundle
+from oic.utils.keyio import KeyBundle, dump_jwks
 from oic.utils.keyio import key_export
 
 from requests import ConnectionError
@@ -37,6 +36,7 @@ from oic.oic.message import DiscoveryResponse
 
 from jwkest import jws, jwe
 from jwkest.jws import alg2keytype
+from jwkest.jws import NoSuitableSigningKeys
 
 __author__ = 'rohe0002'
 
@@ -56,10 +56,10 @@ from oic.utils.http_util import Redirect
 from oic.utils.http_util import BadRequest
 from oic.utils.http_util import Unauthorized
 
-from oic.oauth2 import MissingRequiredAttribute
+from oic.oauth2 import MissingRequiredAttribute, CapabilitiesMisMatch
 from oic.oauth2 import rndstr
 
-from oic.oic import Server
+from oic.oic import Server, PREFERENCE2PROVIDER
 
 from oic.exception import *
 from jwkest.jwe import JWEException
@@ -185,22 +185,88 @@ class Provider(AProvider):
         self.seed = ""
         self.sso_ttl = 0
         self.test_mode = False
+
+        # where the jwks file kan be found by outsiders
         self.jwks_uri = []
+        # Local filename
+        self.jwks_name = ""
 
         self.authn_as = None
         self.preferred_id_type = "public"
         self.hostname = hostname or socket.gethostname
         self.register_endpoint = "%s%s" % (self.baseurl, "register")
 
+        self.jwx_def = {}
+        for _typ in ["sign_alg", "enc_alg", "enc_enc"]:
+            self.jwx_def[_typ] = {}
+            for item in ["request_object", "id_token", "userinfo"]:
+                self.jwx_def[_typ][item] = ""
+
+        self.jwx_def["sign_alg"]["id_token"] = "RS256"
+
+        self.force_jws = {}
+        for item in ["request_object", "id_token", "userinfo"]:
+            self.force_jws[item] = False
+
         if capabilities:
             self.verify_capabilities(capabilities)
-            self.capabilities = capabilities
+            self.capabilities = ProviderConfigurationResponse(**capabilities)
         else:
             self.capabilities = self.provider_features()
+        self.capabilities["issuer"] = self.name
+        self.kid = {"sig": {}, "enc": {}}
 
-    def id_token_as_signed_jwt(self, session, loa="2", alg="RS256", code=None,
+    def set_mode(self, mode):
+        """
+        The mode is a set of parameters that govern how this OP will behave.
+
+        :param mode:
+        :return:
+        """
+
+        # Is there a signing algorithm I should use
+        try:
+            self.jwx_def["sign_alg"]["id_token"] = mode["sign"]
+            self.jwx_def["sign_alg"]["userinfo"] = mode["sign"]
+        except KeyError:
+            pass
+        else:
+        # make sure id_token_signed_response_alg is set in client register
+        # response. This will make it happen in match_preferences()
+            for val in PREFERENCE2PROVIDER.values():
+                if val.endswith("signing_alg_values_supported"):
+                    self.capabilities[val] = [mode["sign"]]
+
+        # Is there a encryption algorithm I should use
+        try:
+            _enc_alg = mode["enc_alg"]
+        except KeyError:
+            pass
+        else:
+        # make sure id_token_signed_response_alg is set in client register
+        # response. This will make it happen in match_preferences()
+            for val in PREFERENCE2PROVIDER.values():
+                if val.endswith("encryption_alg_values_supported"):
+                    self.capabilities[val] = [_enc_alg]
+
+        # Is there a encryption enc algorithm I should use
+        try:
+            _enc_enc = mode["enc_enc"]
+        except KeyError:
+            pass
+        else:
+        # make sure id_token_signed_response_alg is set in client register
+        # response. This will make it happen in match_preferences()
+            for val in PREFERENCE2PROVIDER.values():
+                if val.endswith("encryption_enc_values_supported"):
+                    self.capabilities[val] = [_enc_enc]
+
+    def id_token_as_signed_jwt(self, session, loa="2", alg="", code=None,
                                access_token=None, user_info=None, auth_time=0):
 
+        if alg == "":
+            alg = self.jwx_def["sign_alg"]["id_token"]
+
         logger.debug("Signing alg: %s [%s]" % (alg, alg2keytype(alg)))
         _idt = self.server.make_id_token(session, loa, self.baseurl, alg, code,
                                          access_token, user_info, auth_time)
@@ -215,7 +281,8 @@ class Provider(AProvider):
             if "" in self.keyjar:
                 for b in self.keyjar[""]:
                     logger.debug("OC3 server keys: %s" % b)
-                ckey = self.keyjar.get_signing_key(alg2keytype(alg), "")
+                ckey = self.keyjar.get_signing_key(alg2keytype(alg), "",
+                                                   alg=alg)
             else:
                 ckey = None
         logger.debug("ckey: %s" % ckey)
@@ -303,6 +370,11 @@ class Provider(AProvider):
                 req_user = areq["id_token"]["sub"]
             except KeyError:
                 pass
+            else:
+                try:
+                    assert areq["client_id"] in areq["id_token"]["aud"]
+                except AssertionError:
+                    req_user = ""  # Not allow to use
 
         return req_user
 
@@ -342,13 +414,20 @@ class Provider(AProvider):
                     if not comparision_type:
                         comparision_type = "exact"
 
+                if not isinstance(areq["acr_values"], list):
+                    areq["acr_values"] = [areq["acr_values"]]
+
                 for acr in areq["acr_values"]:
                     res = self.authn_broker.pick(acr, comparision_type)
+                    logger.debug("Picked AuthN broker for ACR %s: %s" % (
+                        str(acr), str(res)))
                     if res:
                         #Return the best guess by pick.
                         return res[0]
-        except KeyError:
-            pass
+        except KeyError as exc:
+            logger.debug(
+                "An error occured while picking the authN broker: %s" % str(
+                    exc))
 
         # return the best I have
         return None, None
@@ -361,8 +440,10 @@ class Provider(AProvider):
             client_info = self.cdb[self.sdb.getClient_id(uid)]
             if redirect_uri in client_info["post_logout_redirect_uris"]:
                 return redirect_uri
-        except:
-            pass
+        except Exception as exc:
+            logger.debug(
+                "An error occurred while verifying redir URI: %s" % str(exc))
+
         return None
 
     def is_session_revoked(self, request="", cookie=None):
@@ -399,7 +480,8 @@ class Provider(AProvider):
         esr = EndSessionRequest().from_urlencoded(request)
         redirect_uri = self.verify_post_logout_redirect_uri(esr, cookie)
         if not redirect_uri:
-            return self._error_response("Not allowed!")
+            return self._error_response(
+                "Not allowed (Post logout redirect URI verification failed)!")
 
         authn, acr = self.pick_auth(esr)
 
@@ -413,7 +495,8 @@ class Provider(AProvider):
             try:
                 uid = identity["uid"]
             except KeyError:
-                return self._error_response("Not allowed!")
+                return self._error_response(
+                    "Not allowed (UID could not be retrieved)!")
 
         #if self.sdb.get_verified_logout(uid):
         #    return self.let_user_verify_logout(uid, esr, cookie, redirect_uri)
@@ -461,6 +544,23 @@ class Provider(AProvider):
         kwargs["cookie"] = cookie
         return authn.verify(_req, **kwargs)
 
+    def setup_session(self, areq, authn_event, cinfo):
+        try:
+            oidc_req = areq["request"]
+        except KeyError:
+            oidc_req = None
+
+        sid = self.sdb.create_authz_session(authn_event, areq, oidreq=oidc_req)
+        kwargs = {}
+        for param in ["sector_id", "preferred_id_type"]:
+            try:
+                kwargs[param] = cinfo[param]
+            except KeyError:
+                pass
+
+        self.sdb.do_sub(sid, **kwargs)
+        return sid
+
     def authorization_endpoint(self, request="", cookie=None, **kwargs):
         """ The AuthorizationRequest endpoint
 
@@ -505,9 +605,24 @@ class Provider(AProvider):
                                               "%s" % err)
 
         areq = self.handle_oidc_request(areq, redirect_uri)
+        if isinstance(areq, Response):
+            return areq
         logger.debug("AuthzRequest+oidc_request: %s" % (areq.to_dict(),))
 
+        cinfo = self.cdb[areq["client_id"]]
         req_user = self.required_user(areq)
+        if req_user:
+            try:
+                sids = self.sdb.sub2sid[req_user]
+            except KeyError:
+                pass
+            else:
+                # anyone will do
+                authn_event = self.sdb[sids[0]]["authn_event"]
+                # Is the authentication event to be regarded as valid ?
+                if authn_event.valid():
+                    sid = self.setup_session(areq, authn_event, cinfo)
+                    return self.authz_part2(authn_event.uid, areq, sid)
 
         authn, authn_class_ref = self.pick_auth(areq)
         if not authn:
@@ -515,7 +630,6 @@ class Provider(AProvider):
             if not authn:
                 authn, authn_class_ref = self.pick_auth(areq, "any")
 
-        logger.debug("Cookie: %s" % cookie)
         try:
             try:
                 _auth_info = kwargs["authn"]
@@ -524,7 +638,6 @@ class Provider(AProvider):
             identity = authn.authenticated_as(cookie,
                                               authorization=_auth_info,
                                               max_age=self.max_age(areq))
-            auth_time = time_util.utc_now()
         except (NoSuchAuthentication, ToOld, TamperAllert):
             identity = None
 
@@ -533,7 +646,6 @@ class Provider(AProvider):
                       "as_user": req_user,
                       "authn_class_ref": authn_class_ref}
 
-        cinfo = self.cdb[areq["client_id"]]
         for attr in ["policy_uri", "logo_uri"]:
             try:
                 authn_args[attr] = cinfo[attr]
@@ -564,16 +676,12 @@ class Provider(AProvider):
                     else:
                         return authn(**authn_args)
 
+        authn_event = AuthnEvent(identity["uid"], authn_info=authn_class_ref)
+
         logger.debug("- authenticated -")
         logger.debug("AREQ keys: %s" % areq.keys())
 
-        try:
-            oidc_req = areq["request"]
-        except KeyError:
-            oidc_req = None
-
-        sid = self.sdb.create_authz_session(user, areq, oidreq=oidc_req,
-                                            auth_time=auth_time)
+        sid = self.setup_session(areq, authn_event, cinfo)
         return self.authz_part2(user, areq, sid)
 
     def userinfo_in_id_token_claims(self, session):
@@ -589,13 +697,14 @@ class Provider(AProvider):
         _claims = by_schema(OpenIDSchema, **itc)
 
         if _claims:
-            return self._collect_user_info(session, {"claims": _claims})
+            return self._collect_user_info(session, _claims)
         else:
             return None
 
     def encrypt(self, payload, client_info, cid, val_type="id_token"):
         """
-        Handles the encryption of a payload
+        Handles the encryption of a payload.
+        Shouldn't get here unless there are encrypt parameters in client info
 
         :param payload: The information to be encrypted
         :param client_info: Client information
@@ -603,11 +712,12 @@ class Provider(AProvider):
         :return: The encrypted information as a JWT
         """
 
-        alg = client_info["%s_encrypted_response_alg" % val_type]
         try:
+            alg = client_info["%s_encrypted_response_alg" % val_type]
             enc = client_info["%s_encrypted_response_enc" % val_type]
-        except KeyError:
-            enc = "A128CBC"
+        except KeyError, err:  # both must be defined
+            logger.warning("undefined parameter: %s" % err)
+            raise JWEException("%s undefined" % err)
 
         keys = self.keyjar.get_encrypt_key(owner=cid)
         logger.debug("keys for %s: %s" % (
@@ -636,7 +746,7 @@ class Provider(AProvider):
         try:
             alg = client_info["id_token_signed_response_alg"]
         except KeyError:
-            alg = "RS256"
+            alg = self.jwx_def["sign_alg"]["id_token"]
 
         id_token = self.id_token_as_signed_jwt(sinfo, alg=alg,
                                                code=code,
@@ -688,9 +798,15 @@ class Provider(AProvider):
 
         if "openid" in _info["scope"]:
             userinfo = self.userinfo_in_id_token_claims(_info)
-            _idtoken = self.sign_encrypt_id_token(_info, client_info, req,
-                                                  user_info=userinfo,
-                                                  auth_time=_info["auth_time"])
+            try:
+                _idtoken = self.sign_encrypt_id_token(
+                    _info, client_info, req, user_info=userinfo,
+                    auth_time=_info["authn_event"].authn_time)
+            except (JWEException, NoSuitableSigningKeys) as err:
+                logger.warning(str(err))
+                return self._error(error="access_denied",
+                                   descr="Could not sign/encrypt id_token")
+
             _sdb.update_by_token(_access_code, "id_token", _idtoken)
 
         _log_debug("_tinfo: %s" % _tinfo)
@@ -713,9 +829,9 @@ class Provider(AProvider):
 
         if "openid" in _info["scope"]:
             userinfo = self.userinfo_in_id_token_claims(_info)
-            _idtoken = self.sign_encrypt_id_token(_info, client_info, req,
-                                                  user_info=userinfo,
-                                                  auth_time=_info["auth_time"])
+            _idtoken = self.sign_encrypt_id_token(
+                _info, client_info, req, user_info=userinfo,
+                auth_time=_info["authn_event"].authn_time)
             sid = _sdb.token.get_key(rtoken)
             _sdb.update(sid, "id_token", _idtoken)
 
@@ -752,7 +868,7 @@ class Provider(AProvider):
             client_id = ""
 
         if not client_id:
-            err = TokenErrorResponse(error="unathorized_client")
+            err = TokenErrorResponse(error="unauthorized_client")
             return Unauthorized(err.to_json(), content="application/json")
 
         if not "client_id" in req:  # Optional for access token request
@@ -789,23 +905,12 @@ class Provider(AProvider):
                     pass
 
             if "oidreq" in session:
-                oidreq = OpenIDRequest().deserialize(session["oidreq"], "json")
-                logger.debug("OIDREQ: %s" % oidreq.to_dict())
-                try:
-                    _claims = oidreq["claims"]["userinfo"]
-                except KeyError:
-                    pass
-                else:
-                    for key, val in uic.items():
-                        if key not in _claims:
-                            _claims[key] = val
-                    uic = _claims
-
-                if uic:
-                    userinfo_claims = Claims(**uic)
-                else:
-                    userinfo_claims = None
-            elif uic:
+                uic = self.server.update_claims(session, "oidreq", "userinfo",
+                                                uic)
+            else:
+                uic = self.server.update_claims(session, "authzreq", "userinfo",
+                                                uic)
+            if uic:
                 userinfo_claims = Claims(**uic)
             else:
                 userinfo_claims = None
@@ -813,13 +918,43 @@ class Provider(AProvider):
             logger.debug("userinfo_claim: %s" % userinfo_claims.to_dict())
 
         logger.debug("Session info: %s" % session)
-        info = self.userinfo(session["local_sub"], userinfo_claims)
+        info = self.userinfo(session["authn_event"].uid, userinfo_claims)
 
         info["sub"] = session["sub"]
         logger.debug("user_info_response: %s" % (info,))
 
         return info
 
+    def signed_userinfo(self, client_info, userinfo, session):
+        """
+        Will create a JWS with the userinfo as payload.
+
+        :param client_info: Client registration information
+        :param userinfo: An OpenIDSchema instance
+        :param session: Session information
+        :return: A JWS containing the userinfo as a JWT
+        """
+        try:
+            algo = client_info["userinfo_signed_response_alg"]
+        except KeyError:  # Fall back to default
+            algo = self.jwx_def["sign_alg"]["userinfo"]
+
+        if algo == "none":
+            key = []
+        else:
+            # Use my key for signing
+            key = self.keyjar.get_signing_key(alg2keytype(algo), "", alg=algo)
+            if not key:
+                return self._error(error="access_denied",
+                                   descr="Missing signing key")
+
+        jinfo = userinfo.to_jwt(key, algo)
+        if "userinfo_encrypted_response_alg" in client_info:
+            # encrypt with clients public key
+            jinfo = self.encrypt(jinfo, client_info, session["client_id"],
+                                 "userinfo")
+        return jinfo
+
     #noinspection PyUnusedLocal
     def userinfo_endpoint(self, request="", **kwargs):
         """
@@ -866,18 +1001,8 @@ class Provider(AProvider):
         _cinfo = self.cdb[session["client_id"]]
         try:
             if "userinfo_signed_response_alg" in _cinfo:
-                algo = _cinfo["userinfo_signed_response_alg"]
-                # Use my key for signing
-                key = self.keyjar.get_signing_key(alg2keytype(algo), "")
-                if not key:
-                    return self._error(error="access_denied",
-                                       descr="Missing signing key")
-                jinfo = info.to_jwt(key, algo)
+                jinfo = self.signed_userinfo(_cinfo, info, session)
                 content_type = "application/jwt"
-                if "userinfo_encrypted_response_alg" in _cinfo:
-                    # encrypt with clients public key
-                    jinfo = self.encrypt(jinfo, _cinfo, session["client_id"],
-                                         "userinfo")
             elif "userinfo_encrypted_response_alg" in _cinfo:
                 jinfo = self.encrypt(info.to_json(), _cinfo, session["client_id"],
                                      "userinfo")
@@ -928,6 +1053,19 @@ class Provider(AProvider):
 
         return False
 
+    def match_client_request(self, request):
+        for _pref, _prov in PREFERENCE2PROVIDER.items():
+            if _pref in request:
+                if isinstance(request[_pref], basestring):
+                    try:
+                        assert request[_pref] in self.capabilities[_prov]
+                    except AssertionError:
+                        raise CapabilitiesMisMatch(_pref)
+                else:
+                    if not set(request[_pref]).issubset(
+                            set(self.capabilities[_prov])):
+                        raise CapabilitiesMisMatch(_pref)
+
     def do_client_registration(self, request, client_id, ignore=None):
         if ignore is None:
             ignore = []
@@ -958,14 +1096,43 @@ class Provider(AProvider):
 
         if "redirect_uris" in request:
             ruri = []
+            client_type = request["application_type"]
+
+            if client_type == "web":
+                if request["response_types"] == ["code"]:
+                    must_https = False
+                else:  # one has to be implicit or hybrid
+                    must_https = True
+            else:
+                must_https = False
+
             for uri in request["redirect_uris"]:
-                if urlparse.urlparse(uri).fragment:
+                p = urlparse.urlparse(uri)
+                err = None
+                if client_type == "native" and p.scheme == "http":
+                    if p.hostname != "localhost":
+                        err = ClientRegistrationErrorResponse(
+                            error="invalid_configuration_parameter",
+                            error_description="Http redirect_uri must use localhost")
+                elif must_https and p.scheme != "https":
+                    err = ClientRegistrationErrorResponse(
+                        error="invalid_configuration_parameter",
+                        error_description="None https redirect_uri not allowed")
+                elif p.fragment:
                     err = ClientRegistrationErrorResponse(
                         error="invalid_configuration_parameter",
                         error_description="redirect_uri contains fragment")
+                # This rule will break local testing.
+                # elif must_https and p.hostname == "localhost":
+                #     err = ClientRegistrationErrorResponse(
+                #         error="invalid_configuration_parameter",
+                #         error_description="https redirect_uri with host localhost")
+
+                if err:
                     return Response(err.to_json(),
                                     content="application/json",
                                     status="400 Bad Request")
+
                 base, query = urllib.splitquery(uri)
                 if query:
                     ruri.append((base, urlparse.parse_qs(query)))
@@ -1037,6 +1204,19 @@ class Provider(AProvider):
                         "invalid_configuration_parameter",
                         descr="%s pointed to illegal URL" % item)
 
+        # Do I have the necessary keys
+        for item in ["id_token_signed_response_alg",
+                     "userinfo_signed_response_alg"]:
+            if item in request:
+                if request[item] in self.capabilities[PREFERENCE2PROVIDER[item]]:
+                    ktyp = jws.alg2keytype(request[item])
+                    # do I have this ktyp and for EC type keys the curve
+                    if ktyp not in ["none", "OCT"]:
+                        _k = self.keyjar.get_signing_key(ktyp,
+                                                         alg=request[item])
+                        if not _k:
+                            del _cinfo[item]
+
         try:
             self.keyjar.load_keys(request, client_id)
             try:
@@ -1057,32 +1237,19 @@ class Provider(AProvider):
         return _cinfo
 
     @staticmethod
-    def comb_post_logout_redirect_uris(args):
-        if "post_logout_redirect_uris" not in args:
-            return
+    def comb_uri(args):
+        for param in ["redirect_uris", "post_logout_redirect_uris"]:
+            if param not in args:
+                continue
 
-        val = []
-        for base, query in args["post_logout_redirect_uris"]:
-            if query:
-                val.append("%s?%s" % (base, query))
-            else:
-                val.append(base)
+            val = []
+            for base, query in args[param]:
+                if query:
+                    val.append("%s?%s" % (base, query))
+                else:
+                    val.append(base)
 
-        args["post_logout_redirect_uris"] = val
-
-    @staticmethod
-    def comb_redirect_uris(args):
-        if "redirect_uris" not in args:
-            return
-
-        val = []
-        for base, query in args["redirect_uris"]:
-            if query:
-                val.append("%s?%s" % (base, query))
-            else:
-                val.append(base)
-
-        args["redirect_uris"] = val
+            args[param] = val
 
     #noinspection PyUnusedLocal
     def l_registration_endpoint(self, request, authn=None, **kwargs):
@@ -1109,6 +1276,12 @@ class Provider(AProvider):
                 return self._error(error="invalid_configuration_parameter",
                                    descr="%s" % err)
 
+        try:
+            self.match_client_request(request)
+        except CapabilitiesMisMatch, err:
+            return self._error(error="invalid_request",
+                               descr="Don't support proposed %s" % err)
+
         _keyjar = self.server.keyjar
 
         # create new id och secret
@@ -1145,8 +1318,7 @@ class Provider(AProvider):
         args = dict([(k, v) for k, v in _cinfo.items()
                      if k in RegistrationResponse.c_param])
 
-        self.comb_redirect_uris(args)
-        self.comb_post_logout_redirect_uris(args)
+        self.comb_uri(args)
         response = RegistrationResponse(**args)
 
         # Add the client_secret as a symmetric key to the keyjar
@@ -1200,7 +1372,7 @@ class Provider(AProvider):
         args = dict([(k, v) for k, v in self.cdb[client_id].items()
                      if k in RegistrationResponse.c_param])
 
-        self.comb_redirect_uris(args)
+        self.comb_uri(args)
         response = RegistrationResponse(**args)
 
         return Response(response.to_json(), content="application/json",
@@ -1215,10 +1387,13 @@ class Provider(AProvider):
         :return:
         """
 
-        _provider_info = copy.deepcopy(self.capabilities)
+        if not self.baseurl.endswith("/"):
+            self.baseurl += "/"
+
+        _provider_info = self.capabilities
 
         if setup:
-            # sort of placeholder tight now
+            # sort of placeholder right now
             # for key, spec in pcr_class.c_param.items():
             #     if isinstance(setup[0], list) and setup[0] != basestring:
             #         # dealing with lists
@@ -1243,9 +1418,6 @@ class Provider(AProvider):
         _provider_info["issuer"] = self.baseurl
         _provider_info["version"] = "3.0"
 
-        if not self.baseurl.endswith("/"):
-            self.baseurl += "/"
-
         return _provider_info
 
     def provider_features(self, pcr_class=ProviderConfigurationResponse):
@@ -1255,11 +1427,19 @@ class Provider(AProvider):
         :param pcr_class:
         :return: ProviderConfigurationResponse instance
         """
+        _scopes = SCOPE2CLAIMS.keys()
+        _scopes.append("openid")
+
+        _claims = []
+        for _cl in SCOPE2CLAIMS.values():
+            _claims.extend(_cl)
+        _claims = list(set(_claims))
+
         _provider_info = pcr_class(
             token_endpoint_auth_methods_supported=[
                 "client_secret_post", "client_secret_basic",
                 "client_secret_jwt", "private_key_jwt"],
-            scopes_supported=["openid"],
+            scopes_supported=_scopes,
             response_types_supported=["code", "token", "id_token",
                                       "code token", "code id_token",
                                       "token id_token",
@@ -1270,7 +1450,7 @@ class Provider(AProvider):
                 "authorization_code", "implicit",
                 "urn:ietf:params:oauth:grant-type:jwt-bearer"],
             claim_types_supported=["normal", "aggregated", "distributed"],
-            claims_supported=SCOPE2CLAIMS.keys(),
+            claims_supported=_claims,
             claims_parameter_supported="true",
             request_parameter_supported="true",
             request_uri_parameter_supported="true",
@@ -1308,7 +1488,7 @@ class Provider(AProvider):
         :return: True or False
         """
         _pinfo = self.provider_features()
-        for key, val in capabilities:
+        for key, val in capabilities.items():
             if isinstance(val, basestring):
                 try:
                     if val in _pinfo[key]:
@@ -1391,7 +1571,6 @@ class Provider(AProvider):
 
         # Do the authorization
         try:
-            info = OpenIDSchema(**self._collect_user_info(self.sdb[sid]))
             permission = self.authz(user)
             self.sdb.update(sid, "permission", permission)
         except Exception:
@@ -1460,10 +1639,18 @@ class Provider(AProvider):
                 user_info = self.userinfo_in_id_token_claims(_sinfo)
                 client_info = self.cdb[areq["client_id"]]
 
+                hargs = {}
+                if set(areq["response_type"]) == {'code', 'id_token', 'token'}:
+                    hargs = {"code": _code, "access_token": _access_token}
+                elif set(areq["response_type"]) == {'code', 'id_token'}:
+                    hargs = {"code": _code}
+                elif set(areq["response_type"]) == {'id_token', 'token'}:
+                    hargs = {"access_token": _access_token}
+
+                # or 'code id_token'
                 id_token = self.sign_encrypt_id_token(
-                    _sinfo, client_info, areq, code=_code,
-                    access_token=_access_token, user_info=user_info,
-                    auth_time=_sinfo["auth_time"])
+                    _sinfo, client_info, areq, user_info=user_info,
+                    auth_time=_sinfo["authn_event"].authn_time, **hargs)
 
                 aresp["id_token"] = id_token
                 _sinfo["id_token"] = id_token
@@ -1538,6 +1725,57 @@ class Provider(AProvider):
         """
         return self.end_session_endpoint(request, **kwargs)
 
+    def do_key_rollover(self, jwks, kid_template):
+        """
+        Handle key roll-over by importing new keys and inactivating the
+        ones in the keyjar that are of the same type and usage.
+
+        :param jwk: A JWK
+        """
+
+        kb = KeyBundle()
+        kb.do_keys(jwks["keys"])
+
+        kid = 0
+        for k in kb.keys():
+            if not k.kid:
+                k.kid = kid_template % kid
+                kid += 1
+            self.kid[k.use][k.kty] = k.kid
+
+            # find the old key for this key type and usage and mark that
+            # as inactive
+            for _kb in self.keyjar.issuer_keys[""]:
+                for key in _kb.keys():
+                    if key.kty == k.kty and key.use == k.use:
+                        if k.kty == "EC":
+                            if key.crv == k.crv:
+                                key.inactive_since = time.time()
+                        else:
+                            key.inactive_since = time.time()
+
+        self.keyjar.add_kb("", kb)
+
+        if self.jwks_name:
+            # print to the jwks file
+            dump_jwks(self.keyjar[""], self.jwks_name)
+
+    def remove_inactive_keys(self, more_then=3600):
+        """
+        Remove all keys that has been inactive 'more_then' seconds
+
+        :param more_then: An integer (default = 3600 seconds == 1 hour)
+        """
+        now = time.time()
+        for kb in self.keyjar.issuer_keys[""]:
+            for key in kb.keys():
+                if key.inactive_since:
+                    if now - key.inactive_since > more_then:
+                        kb.remove(key)
+            if len(kb) == 0:
+                self.keyjar.issuer_keys[""].remove(kb)
+
+
 # -----------------------------------------------------------------------------
 
 
