@@ -169,13 +169,6 @@ class BearerHeader(ClientAuthnMethod):
                     try:
                         _acc_token = kwargs["access_token"]
                     except KeyError:
-                        try:
-                            _ = kwargs["state"]
-                        except KeyError:
-                            if not self.cli.state:
-                                raise Exception("Missing state specification")
-                            kwargs["state"] = self.cli.state
-
                         _acc_token = self.cli.get_token(**kwargs).access_token
         else:
             try:
@@ -264,7 +257,8 @@ class JWSAuthnMethod(ClientAuthnMethod):
         return algorithm
 
     def get_signing_key(self, algorithm):
-        return self.cli.keyjar.get_signing_key(alg2keytype(algorithm))
+        return self.cli.keyjar.get_signing_key(alg2keytype(algorithm),
+                                               alg=algorithm)
 
     def get_key_by_kid(self, kid, algorithm):
         _key = self.cli.keyjar.get_key_by_kid(kid)
@@ -331,10 +325,15 @@ class JWSAuthnMethod(ClientAuthnMethod):
             return False
 
         logger.debug("authntoken: %s" % bjwt.to_dict())
-        logger.debug("known clients: %s" % self.cli.cdb.keys())
+        #logger.debug("known clients: %s" % self.cli.cdb.keys())
+        try:
+            cid = kwargs["client_id"]
+        except KeyError:
+            cid = bjwt["iss"]
+
         try:
             # There might not be a client_id in the request
-            assert str(bjwt["iss"]) in self.cli.cdb  # It's a client I know
+            assert str(cid) in self.cli.cdb  # It's a client I know
         except KeyError:
             pass
 
@@ -352,7 +351,7 @@ class JWSAuthnMethod(ClientAuthnMethod):
         except AssertionError:
             raise NotForMe("Not for me!")
 
-        return True
+        return cid
 
 
 class ClientSecretJWT(JWSAuthnMethod):
@@ -366,7 +365,8 @@ class ClientSecretJWT(JWSAuthnMethod):
         return JWSAuthnMethod.choose_algorithm(self, entity, **kwargs)
 
     def get_signing_key(self, algorithm):
-        return self.cli.keyjar.get_signing_key(alg2keytype(algorithm))
+        return self.cli.keyjar.get_signing_key(alg2keytype(algorithm),
+                                               alg=algorithm)
 
 
 class PrivateKeyJWT(JWSAuthnMethod):
@@ -377,7 +377,8 @@ class PrivateKeyJWT(JWSAuthnMethod):
         return JWSAuthnMethod.choose_algorithm(self, entity, **kwargs)
 
     def get_signing_key(self, algorithm):
-        return self.cli.keyjar.get_signing_key(alg2keytype(algorithm), "")
+        return self.cli.keyjar.get_signing_key(alg2keytype(algorithm), "",
+                                               alg=algorithm)
 
 
 #from oic.utils.authn.client_saml import SAML2_BEARER_ASSERTION_TYPE
@@ -412,7 +413,7 @@ def verify_client(inst, areq, authn, type_method=TYPE_METHOD):
     elif "client_assertion" in areq:  # client_secret_jwt or public_key_jwt
         for typ, method in type_method:
             if areq["client_assertion_type"] == typ:
-                return method(inst).verify(areq)
+                return method(inst).verify(areq, client_id=client_id)
         else:
             raise UnknownAssertionType(areq["client_assertion_type"], areq)
     else:
