@@ -1,3 +1,5 @@
+import logging
+
 __author__ = 'rohe0002'
 
 import cgi
@@ -15,6 +17,8 @@ from oic.utils.aes import decrypt
 
 from Cookie import SimpleCookie
 
+logger = logging.getLogger(__name__)
+
 
 class Response(object):
     _template = None
@@ -310,26 +314,23 @@ def get_or_post(environ):
     return data
 
 
-def wsgi_wrapper(environ, start_response, func, **kwargs):
+def extract_from_request(environ, kwargs={}):
     request = None
     try:
         request = environ["QUERY_STRING"]
     except KeyError:
         pass
-
     if not request:
         try:
             request = get_post(environ)
         except KeyError:
             pass
-
     kwargs["request"] = request
     # authentication information
     try:
         kwargs["authn"] = environ["HTTP_AUTHORIZATION"]
     except KeyError:
         pass
-
     try:
         kwargs["cookie"] = environ["HTTP_COOKIE"]
     except KeyError:
@@ -340,20 +341,34 @@ def wsgi_wrapper(environ, start_response, func, **kwargs):
     kwargs["url"] = geturl(environ, query=False)
     kwargs["baseurl"] = geturl(environ, query=False, path=False)
     kwargs["path"] = getpath(environ)
+    return kwargs
 
+
+def wsgi_wrapper(environ, start_response, func, **kwargs):
+    kwargs = extract_from_request(environ, kwargs)
     args = func(**kwargs)
+
     try:
-        resp, argv = args
-        return resp(environ, start_response, **argv)
+        resp, state = args
+        return resp(environ, start_response)
     except TypeError:
         resp = args
         return resp(environ, start_response)
     except Exception as err:
-        #logger.error("%s" % err)
+        logger.error("%s" % err)
         raise
 
 
 class CookieDealer(object):
+
+    def getServer(self):
+        return self._srv
+
+    def setServer(self, server):
+        self._srv = server
+
+    srv = property(getServer, setServer)
+
     def __init__(self, srv, ttl=5):
         self.srv = None
         self.init_srv(srv)
